/*
 *  linux/arch/ppc/kernel/traps.c
 *
 *  Copyright (C) 1995-1996  Gary Thomas (gdt@linuxppc.org)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 */

/*
 * This file handles the architecture-dependent parts of hardware exceptions
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/user.h>
#include <linux/a.out.h>
#include <linux/init.h>

#include <asm/pgtable.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/processor.h>
#include <asm/prom.h>
#include <asm/ppcops.h>

extern void debug(char *, struct pt_regs *);
extern void bad_page_fault(struct pt_regs *, unsigned long);
extern void dump_regs(struct pt_regs *regs);
extern void hard_reset_now(void);

extern struct thread_struct *fpu_tss;

/*
 * Trap & Exception support
 */

void __init
trap_init(void)
{
}

void
_exception(int signr, char *msg, struct pt_regs *regs)
{
	if (!user_mode(regs)) {
                debug(msg, regs);
	} else {
		force_sig(signr, current);
	}
}

void
DebugException(struct pt_regs *regs) {
	/*
	 * Should be a lot more powerful. For now, ignore
	 * most debug features of the chip.
	 */
	_exception(SIGTRAP, "Debug Exception", regs);	
}

#ifndef CONFIG_XMON
void
CriticalInterruptException(struct pt_regs *regs) {
	_exception(0, "Critical Interrupt Exception", regs);	
}
#endif

void
MachineCheckException(struct pt_regs *regs)
{
	unsigned fixup;
	if (!current) {
		if (regs->nip >= KERNELBASE
		    && (fixup = search_exception_table(regs->nip)) != 0) {
			regs->nip = fixup;
			return;
		}
	}
	else if ((current->tss.besr & 0x80000000) && !(current->tss.besr & 0x40000000)) {
		unsigned int code = current->tss.besr & 0x1c000000;
		if (code == 0 || code == 4 || code == 6) {
			if (regs->nip >= KERNELBASE
			    && (fixup = search_exception_table(regs->nip)) != 0) {
				regs->nip = fixup;
				return;
			}
		}
	}
	_exception(SIGBUS, "Bus Error", regs);	
}

void
ProgramCheckException(struct pt_regs *regs)
{
	unsigned int instr;
	int fix_fp(struct pt_regs *);
	unsigned int esr = mfspr(SPRN_ESR);

        if (!(esr&0x08000000)) printk("PCE: esr=0x%x, nip=0x%lx, msr=0x%lx, instr=0x%x\n", 
            esr, regs->nip, regs->msr, *((unsigned int *)regs->nip));
	if (esr & 0x02000000) {
		DebugException(regs);
		return;
	}
	if (esr & 0x08000000) {
		if (fix_fp(regs))
			return;
	}

	instr = *((unsigned int *)regs->nip);
	if ((instr & 0xffff0000) == 0x398c0000 ||
	    (instr & 0xffff0000) == 0x7d890000 ||
	    (instr & 0xffff0000) == 0x3d800000) {
                if (current->tss.flags != regs->nip) {
                    current->tss.flags = regs->nip;
		    store_cache_range(regs->nip, regs->nip + 16);
		    printk("cache flush program check (fix attempted) ...\n");
		    return;
		}
                printk("cache flush program check (fix failed!) ...\n");
	}

	sti();
	_exception(SIGILL, "Illegal Instruction", regs);
}

void
AlignmentException(struct pt_regs *regs)
{
	int fix_alignment(struct pt_regs *regs);
	if (fix_alignment(regs) <= 0) {
		_exception(SIGBUS, "Alignment Exception", regs);	
	}
        return;
}

void
UnknownException(struct pt_regs *regs)
{
	_exception(SIGTRAP, "Unknown Exception", regs);	
}

void
bad_stack(struct pt_regs *regs)
{
        _exception(0, "Kernel Stack Overflow", regs);
}

void dump_regs(struct pt_regs *regs) {
        debug("Register dump requested", regs);
}

#ifndef CONFIG_XMON
int
getsp(void)
{
    int x;

    asm("mr %0,1" : "=r" (x) :);
    return x;
}

void
debug(char *s, struct pt_regs *regs) {
        extern int sys_reboot(int, int, int);
        extern int sys_sync(void);
	extern int int_return;
	extern int critical_return;
	extern int syscall_return;
	unsigned int *sp;
	unsigned int stack[2];

        printk("\nKernel Panic: %s\n", s);
	if (regs)
		printk("Exception 0x%lx: NIP: %08lX, MSR: %08lX, FRAME: 0x%p\n",
		       regs->trap, regs->nip, regs->msr, regs);
	printk("ESR 0x%x BESR 0x%x BEAR 0x%x DEAR 0x%x\n",
		mfspr(SPRN_ESR), mfspr(SPRN_BESR), mfspr(SPRN_BEAR),
		mfspr(SPRN_DEAR));
	if (current)
		printk("TASK = %p[%d] '%s'\n", current, current->pid, current->comm);
	if (regs) {
		int i;
		for (i = 0;  i < 32;  i++) {
			if ((i % 8) == 0)
				printk("GPR%02d: ", i);
			printk("%08lX ", regs->gpr[i]);
			if ((i % 8) == 7)
				printk("\n");
		}
	}

        sp = (unsigned int *) getsp();
        for(; sp != 0; sp = (unsigned int *)(stack[0]))
        {
            stack[0] = sp[0];
            stack[1] = sp[1];
            printk("0x%x ", stack[1]);
            if (stack[1] == (unsigned) &int_return ||
                stack[1] == (unsigned) &critical_return ||
		stack[1] == (unsigned) &syscall_return) {
                struct pt_regs *rp = (struct pt_regs *) ((unsigned long)sp + 16);
                printk("\nexception:%lx [%lx] %lx ", rp->trap, (unsigned long)sp + 16,
                       rp->nip);
            }
	    else if (stack[0] <= (unsigned int)sp) {
		    if (stack[0] != 0)
			    printk("(stack corrupted, next sp was 0x%x)", stack[0]);
		    break;
            }
        }
        printk("\n");

        printk("Rebooting system: it's dead!\n");

        // NOTE: We used to call sys_sync() followed by sys_reboot().  However,
        // if we kenel panic, there is no reason to believe that sys_sync()
        // will really work -- in fact, we have one known case where it didn't
        // work (because the free list was corrupt), and our box just hung,
        // leaving the user high and dry (and confused).
        //
        // So, now we just reset.  No messing around, just do it.  Of course,
        // we may not get this crash logged properly, but at least the user
        // gets his TiVo back...
        hard_reset_now();

        // NOT REACHED
        sys_reboot(0xfee1dead, 672274793, 0x1234567);
}
#endif

int do_trace_syscalls;

void
trace_syscall(struct pt_regs *regs)
{
	static int count;
	if (!do_trace_syscalls) return;
	if ((do_trace_syscalls > 0) && (do_trace_syscalls != current->pid)) return;
	printk("Task: %p(%d), PC: %08lX/%08lX, Syscall: %3ld, Result: %s%ld\n",
	       current, current->pid, regs->nip, regs->link, regs->gpr[0],
	       regs->ccr&0x10000000?"Error=":"", regs->gpr[3]);
	if (++count == 20)
	{
		count = 0;
	}
}

#define I_RT(x)		(((x) & 0x03700000) >> 21)
#define I_RA(x)		(((x) & 0x001f0000) >> 16)
#define I_RB(x)		(((x) & 0x0000f100) >> 11)
#define I_D(x)		((x) & 0xffff)
#define I_OP(x)		(((x) & 0x7fe) >> 1)

/*
 * Pretend we have FP registers, and emulate what they would do
 * to the GP registers, without actually doing anything.
 */
int
fix_fp(struct pt_regs *regs) {
	unsigned int instr;
	int code;
	int rval;

	if (!(mfspr(SPRN_ESR) & 0x08000000))
		return 0;
	if (current == NULL)
		return 0;

	instr = *((unsigned int *)regs->nip);
	code = (instr & 0xfc000000) >> 26;
	switch (code) {
	case 0x1f:
		switch (I_OP(instr)) {
		case 0x237:
		case 0x277:
		case 0x2b7:
		case 0x2f7:
			if (I_RA(instr) == 0)
				goto cant;
			rval = regs->gpr[I_RA(instr)];
			rval += (int)regs->gpr[I_RB(instr)];
			regs->gpr[I_RA(instr)] = rval;
		case 0x217:
		case 0x257:
		case 0x297:
		case 0x2d7:
		case 0x3d7:
			regs->nip += 4;
			return 1;
		}
		goto cant;
	case 0x31:
	case 0x33:
	case 0x35:
	case 0x37:
		if (I_RA(instr) == 0)
			goto cant;
		rval = regs->gpr[I_RA(instr)];
		rval += I_D(instr);
		regs->gpr[I_RA(instr)] = rval;
	case 0x30:
	case 0x32:
	case 0x34:
	case 0x36:
		regs->nip += 4;
		return 1;
	case 0x3f:
		switch (I_OP(instr)) {
		case 0xf:		/* fctiwz */
		case 0x48:		/* fmr */
			regs->nip += 4;
			return 1;
		}
	}
    cant:
	return 0;
}
