#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mmu.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include "8390.h"

#define NE_CMD	 	0x00
#define NE_DATAPORT	0x10	/* NatSemi-defined port window offset. */
#define NE_RESET	0x1f	/* Issue a read to reset, a write to clear. */
#define NE_IO_EXTENT	0x40

#define NE1SM_START_PG	0x20	/* First page of TX buffer */
#define NE1SM_STOP_PG 	0x40	/* Last page +1 of RX ring */
#define NESM_START_PG	0x40	/* First page of TX buffer */
#define NESM_STOP_PG	0x80	/* Last page +1 of RX ring */

#ifdef CONFIG_TCD
#ifdef CONFIG_TCD_EVAL
#	define NE_TXPG		NE1SM_START_PG
#	define NE_RXPG		NE1SM_STOP_PG
#else
#	define NE_TXPG		NESM_START_PG
#	define NE_RXPG		128
#endif
#endif

static int tcd_8390_open(struct device *dev);
static int tcd_8390_close(struct device *dev);

static void tcd_8390_reset_8390(struct device *dev);
static void tcd_8390_get_8390_hdr(struct device *dev, struct e8390_pkt_hdr *hdr,
			  int ring_page);
static void tcd_8390_block_input(struct device *dev, int count,
			  struct sk_buff *skb, int ring_offset);
static void tcd_8390_block_output(struct device *dev, const int count,
		const unsigned char *buf, const int start_page);

int tcd_8390_probe(struct device *dev) {
    const char *name = NULL;
    int start_page, stop_page;
    unsigned int ioaddr = IOBASE_8390;
    int reg0;

{
    extern int s403_ready;


   /* If the 16550 UART doesn't respond, assume the mfg board isn't there. */
    if (s403_ready < 0) {
    	dev->base_addr = 0;
	return ENODEV;
    }
}
    dev->base_addr = ioaddr;
{
    /* This is set up early in initialization from the PROM MAC address */
    extern char hw_enet_addr[];

    memcpy(&dev->dev_addr, &hw_enet_addr, 6);
}
    reg0 = inb_p(ioaddr);

    if (reg0 == 0xFF) {
    	dev->base_addr = 0;
	return ENODEV;
    }

    /* Do a preliminary verification that we have a 8390. */
    {	int regd;
	outb_p(E8390_NODMA+E8390_PAGE1+E8390_STOP, ioaddr + E8390_CMD);
	regd = inb_p(ioaddr + 0x0d);
	outb_p(0xff, ioaddr + 0x0d);
	outb_p(E8390_NODMA+E8390_PAGE0, ioaddr + E8390_CMD);
	inb_p(ioaddr + EN0_COUNTER0); /* Clear the counter by reading. */
	if (inb_p(ioaddr + EN0_COUNTER0) != 0) {
	    outb_p(reg0, ioaddr);
	    outb_p(regd, ioaddr + 0x0d);	/* Restore the old values. */
	    dev->base_addr = 0;
	    return ENODEV;
	}
    }

    /* We should have a "dev" from Space.c or the static module table. */
    if (dev == NULL) {
	printk(KERN_ERR "ne.c: Passed a NULL device.\n");
	dev = init_etherdev(0, 0);
    }

    printk("TCD 83902 Ethernet chip at %#3x:", ioaddr);
    name = "TCD8390";
    dev->irq = IRQ_8390;
    start_page = NE_TXPG;
    stop_page = NE_RXPG;

    outb_p(0, ioaddr + EN0_IMR);	/* disable all interrupts for now */
    {
	int irqval = request_irq(dev->irq, ei_interrupt, 0, name, NULL);
	if (irqval) {
	    printk (" unable to get IRQ %d (irqval=%d).\n", dev->irq, irqval);
	    return EAGAIN;
	}
    }

    dev->base_addr = ioaddr;

    /* Allocate dev->priv and fill in 8390 specific dev fields. */
    if (ethdev_init(dev)) {
	printk (" unable to get memory for dev->priv.\n");
	free_irq(dev->irq, NULL);
	return -ENOMEM;
    }

    request_region(ioaddr, (1 * 1024 * 1024), name);

    printk("\n%s: %s found at %#x, using IRQ %d.\n",
	   dev->name, name, ioaddr, dev->irq);

    ei_status.name = name;
    ei_status.tx_start_page = start_page;
    ei_status.stop_page = stop_page;
    ei_status.word16 = 0;	/* only a byte-wide path available */
    ei_status.rx_start_page = start_page + TX_PAGES;
    ei_status.reset_8390 = &tcd_8390_reset_8390;
    ei_status.block_input = &tcd_8390_block_input;
    ei_status.block_output = &tcd_8390_block_output;
    ei_status.get_8390_hdr = &tcd_8390_get_8390_hdr;
    dev->open = &tcd_8390_open;
    dev->stop = &tcd_8390_close;
    NS8390_init(dev, 0);
    return 0;
}

static int
tcd_8390_open(struct device *dev)
{
    ei_open(dev);
    MOD_INC_USE_COUNT;
    return 0;
}

static int
tcd_8390_close(struct device *dev)
{
    if (ei_debug > 1)
	printk("%s: Shutting down ethercard.\n", dev->name);
    ei_close(dev);
    MOD_DEC_USE_COUNT;
    return 0;
}

/* Hard reset the card.  This used to pause for the same period that a
   8390 reset command required, but that shouldn't be necessary. */
static void
tcd_8390_reset_8390(struct device *dev)
{
#ifdef WE_FOUND_OUT_HOW_TO_RESET
    unsigned long reset_start_time = jiffies;

    if (ei_debug > 1) printk("resetting the 8390 t=%ld...", jiffies);

    /* DON'T change these to inb_p/outb_p or reset will fail on clones. */
    outb(inb(NE_BASE + NE_RESET), NE_BASE + NE_RESET);

    ei_status.txing = 0;
    ei_status.dmaing = 0;

    /* This check _should_not_ be necessary, omit eventually. */
    while ((inb_p(NE_BASE+EN0_ISR) & ENISR_RESET) == 0)
	if (jiffies - reset_start_time > 2*HZ/100) {
	    printk("%s: tcd_8390_reset_8390() did not complete.\n", dev->name);
	    break;
	}
    outb_p(ENISR_RESET, NE_BASE + EN0_ISR);	/* Ack intr. */
#else
    printk("Resetting %s ...\n", dev->name);
    outb_p(E8390_STOP|E8390_NODMA|E8390_PAGE0, NE_BASE + E8390_CMD);
    ei_status.txing = 0;
    ei_status.dmaing = 0;
#endif
}

/* Grab the 8390 specific header. Similar to the block_input routine, but
   we don't need to be concerned with ring wrap as the header will be at
   the start of a page, so we optimize accordingly. */

static void
tcd_8390_get_8390_hdr(struct device *dev, struct e8390_pkt_hdr *hdr, int ring_page)
{

    int nic_base = dev->base_addr;

    /* This *shouldn't* happen. If it does, it's the last thing you'll see */
    if (ei_status.dmaing) {
	printk("%s: DMAing conflict in tcd_8390_get_8390_hdr "
	   "[DMAstat:%d][irqlock:%d][intr:%d].\n",
	   dev->name, ei_status.dmaing, ei_status.irqlock,
	   dev->interrupt);
	return;
    }

    ei_status.dmaing |= 0x01;
    outb_p(E8390_NODMA+E8390_PAGE0+E8390_START, nic_base+ NE_CMD);
    outb_p(sizeof(struct e8390_pkt_hdr), nic_base + EN0_RCNTLO);
    outb_p(0, nic_base + EN0_RCNTHI);
    outb_p(0, nic_base + EN0_RSARLO);		/* On page boundary */
    outb_p(ring_page, nic_base + EN0_RSARHI);
    outb_p(E8390_RREAD+E8390_START, nic_base + NE_CMD);

    if (ei_status.word16)
	insw(NE_BASE + NE_DATAPORT, hdr, sizeof(struct e8390_pkt_hdr)>>1);
    else
	insb(NE_BASE + NE_DATAPORT, hdr, sizeof(struct e8390_pkt_hdr));

    outb_p(ENISR_RDC, nic_base + EN0_ISR);	/* Ack intr. */
    ei_status.dmaing &= ~0x01;

	hdr->count=le16_to_cpu(hdr->count);
}

/* Block input and output, similar to the Crynwr packet driver.  If you
   are porting to a new ethercard, look at the packet driver source for hints.
   The NEx000 doesn't share the on-board packet memory -- you have to put
   the packet out through the "remote DMA" dataport using outb. */

static void
tcd_8390_block_input(struct device *dev, int count, struct sk_buff *skb, int ring_offset)
{
    int nic_base = dev->base_addr;
    char *buf = skb->data;
#ifdef CONFIG_TCD_EVAL
    unsigned long cpuflags;
#endif

    /* This *shouldn't* happen. If it does, it's the last thing you'll see */
    if (ei_status.dmaing) {
	printk("%s: DMAing conflict in tcd_8390_block_input "
	   "[DMAstat:%d][irqlock:%d][intr:%d].\n",
	   dev->name, ei_status.dmaing, ei_status.irqlock,
	   dev->interrupt);
	return;
    }

#ifdef CONFIG_TCD_EVAL
    save_flags(cpuflags);
    cli();
#endif

    ei_status.dmaing |= 0x01;
    outb_p(E8390_NODMA+E8390_PAGE0+E8390_START, nic_base+ NE_CMD);
    outb_p(count & 0xff, nic_base + EN0_RCNTLO);
    outb_p(count >> 8, nic_base + EN0_RCNTHI);
    outb_p(ring_offset & 0xff, nic_base + EN0_RSARLO);
    outb_p(ring_offset >> 8, nic_base + EN0_RSARHI);
    outb_p(E8390_RREAD+E8390_START, nic_base + NE_CMD);
    if (ei_status.word16) {
      insw(NE_BASE + NE_DATAPORT,buf,count>>1);
      if (count & 0x01) {
	buf[count-1] = inb(NE_BASE + NE_DATAPORT);
      }
    } else {
	insb(NE_BASE + NE_DATAPORT, buf, count);
    }
    outb_p(ENISR_RDC, nic_base + EN0_ISR);	/* Ack intr. */
    ei_status.dmaing &= ~0x01;

#ifdef CONFIG_TCD_EVAL
    restore_flags(cpuflags);
#endif
}

static void
tcd_8390_block_output(struct device *dev, int count,
		const unsigned char *buf, const int start_page)
{
    int nic_base = NE_BASE;
    int bugwhack;
    unsigned long dma_start;
    unsigned char lobyte;
#ifdef CONFIG_TCD_EVAL
    unsigned long cpuflags;
#endif

    /* Round the count up for word writes.  Do we need to do this?
       What effect will an odd byte count have on the 8390?
       I should check someday. */
    if (ei_status.word16 && (count & 0x01))
      count++;

    /* This *shouldn't* happen. If it does, it's the last thing you'll see */
    if (ei_status.dmaing) {
	printk("%s: DMAing conflict in tcd_8390_block_output."
	   "[DMAstat:%d][irqlock:%d][intr:%d]\n",
	   dev->name, ei_status.dmaing, ei_status.irqlock,
	   dev->interrupt);
	return;
    }

#ifdef CONFIG_TCD_EVAL
    save_flags(cpuflags);
    cli();
#endif

    ei_status.dmaing |= 0x01;
    /* We should already be in page 0, but to be safe... */
    outb_p(E8390_PAGE0+E8390_START+E8390_NODMA, nic_base + NE_CMD);

    /* hmmmm. The 83902 documentation states that the processor needs
       to do a "dummy read" before doing the remote write to work around
       a chip bug they don't feel like fixing. So do that. */
    bugwhack = 0;
    while (1) {
	unsigned int rdhi;
	unsigned int rdlo;
	/* Now the normal output. */
	outb_p(ENISR_RDC, nic_base + EN0_ISR);
	outb_p(count & 0xff, nic_base + EN0_RCNTLO);
	outb_p(count >> 8,   nic_base + EN0_RCNTHI);
	outb_p(0x00, nic_base + EN0_RSARLO);
	outb_p(start_page, nic_base + EN0_RSARHI);

	if (bugwhack++)
		break;

	/* do the dummy read */
	rdhi = inb_p(nic_base + EN0_CRDAHI);
	rdlo = inb_p(nic_base + EN0_CRDALO);
	outb_p(E8390_RREAD+E8390_START, nic_base + NE_CMD);
	while (1) {
		unsigned int nrdhi;
		unsigned int nrdlo;
		nrdhi = inb_p(nic_base + EN0_CRDAHI);
		nrdlo = inb_p(nic_base + EN0_CRDALO);
		if ((rdhi != nrdhi) || (rdlo != nrdlo))
			break;
	}
    }
    outb_p(E8390_RWRITE+E8390_START, nic_base + NE_CMD);
    if (ei_status.word16) {
	outsw(NE_BASE + NE_DATAPORT, buf, count>>1);
    } else {
	outsb(NE_BASE + NE_DATAPORT, buf, count);
    }

#ifdef CONFIG_TCD_EVAL
    restore_flags(cpuflags);
#endif
    dma_start = jiffies;

    while (((lobyte = inb_p(nic_base + EN0_ISR)) & ENISR_RDC) == 0)
	if (jiffies - dma_start > 2*HZ/100) {		/* 20ms */
		unsigned char hicnt, locnt;
		hicnt = inb_p(nic_base + EN0_CRDAHI);
		locnt = inb_p(nic_base + EN0_CRDALO);
		printk("%s: timeout waiting for Tx RDC, stat = 0x%x\n",
			dev->name, lobyte);
		printk("\tstart address 0x%x, current address 0x%x, count %d\n",
			(start_page << 8), (hicnt << 8) | locnt, count);
		tcd_8390_reset_8390(dev);
		NS8390_init(dev,1);
		break;
	}

    outb_p(ENISR_RDC, nic_base + EN0_ISR);	/* Ack intr. */
    ei_status.dmaing &= ~0x01;
    return;
}


#ifdef MODULE
#define MAX_NE_CARDS	4	/* Max number of NE cards per module */
#define NAMELEN		8	/* # of chars for storing dev->name */
static char namelist[NAMELEN * MAX_NE_CARDS] = { 0, };
static struct device dev_ne[MAX_NE_CARDS] = {
	{
		NULL,		/* assign a chunk of namelist[] below */
		0, 0, 0, 0,
		0, 0,
		0, 0, 0, NULL, NULL
	},
};

static int io[MAX_NE_CARDS] = { 0, };
static int irq[MAX_NE_CARDS]  = { 0, };
static int bad[MAX_NE_CARDS]  = { 0, };	/* 0xbad = bad sig or no reset ack */

MODULE_PARM(io, "1-" __MODULE_STRING(MAX_NE_CARDS) "i");
MODULE_PARM(irq, "1-" __MODULE_STRING(MAX_NE_CARDS) "i");
MODULE_PARM(bad, "1-" __MODULE_STRING(MAX_NE_CARDS) "i");

/* This is set up so that no ISA autoprobe takes place. We can't guarantee
that the ne2k probe is the last 8390 based probe to take place (as it
is at boot) and so the probe will get confused by any other 8390 cards.
ISA device autoprobes on a running machine are not recommended anyway. */

int
init_module(void)
{
	int this_dev, found = 0;

	for (this_dev = 0; this_dev < MAX_NE_CARDS; this_dev++) {
		struct device *dev = &dev_ne[this_dev];
		dev->name = namelist+(NAMELEN*this_dev);
		dev->irq = irq[this_dev];
		dev->mem_end = bad[this_dev];
		dev->base_addr = io[this_dev];
		dev->init = tcd_8390_probe;
		if (register_netdev(dev) == 0) {
			found++;
			continue;
		}
		if (found != 0) 	/* Got at least one. */
			return 0;
		if (io[this_dev] != 0)
			printk(KERN_WARNING "ne.c: No NE*000 card found at i/o = %#x\n", io[this_dev]);
		else
			printk(KERN_NOTICE "ne.c: No PCI cards found. Use \"io=0xNNN\" value(s) for ISA cards.\n");
		return -ENXIO;
	}

	return 0;
}

void
cleanup_module(void)
{
	int this_dev;

	for (this_dev = 0; this_dev < MAX_NE_CARDS; this_dev++) {
		struct device *dev = &dev_ne[this_dev];
		if (dev->priv != NULL) {
			kfree(dev->priv);
			dev->priv = NULL;
			free_irq(dev->irq, NULL);
			irq2dev_map[dev->irq] = NULL;
			release_region(dev->base_addr, NE_IO_EXTENT);
			unregister_netdev(dev);
		}
	}
}
#endif /* MODULE */

/*
 * Local variables:
 *  compile-command: "make -C../.. modules"
 *  version-control: t
 *  kept-new-versions: 5
 *  c-indent-level: 8
 *  c-brace-imaginary-offset: 0
 *  c-brace-offset: -8
 *  c-argdecl-indent: 8
 *  c-label-offset: -8
 *  c-continued-statement-offset: 8
 *  c-continued-brace-offset: 0
 *  tab-width: 4
 *  indent-tabs-mode: t
 * End:
 */
